from .audio import Audio, ChapterMark, sync_chapter_marks
from .file import File
from .gallery import Gallery, get_or_create_gallery
from .itunes import ItunesArtWork
from .moderation import SpamFilter
from .pages import Blog, HomePage, Post, sync_media_ids
from .request import Request
from .video import Video, get_video_dimensions


__all__ = [
    Audio,
    ChapterMark,
    sync_chapter_marks,
    Blog,
    File,
    Gallery,
    get_or_create_gallery,
    HomePage,
    ItunesArtWork,
    Post,
    sync_media_ids,
    Request,
    SpamFilter,
    Video,
    get_video_dimensions,
]



from django.urls import reverse

import pytest

from .factories import UserFactory


class TestDashboard:
    pytestmark = pytest.mark.django_db

    @classmethod
    def setup_class(cls):
        cls.dashboard_url = reverse("cast:dashboard")

    def test_get_dashboard_without_authentication(self, client):
        r = client.get(self.dashboard_url)
        # redirect to login page
        assert r.status_code == 302

    def test_get_dashboard_with_authentication(self, client):
        user = UserFactory()
        client.login(username=user.username, password="password")
        r = client.get(self.dashboard_url)
        # dont redirect to login page
        assert r.status_code == 200
        assert "Dashboard" in r.content.decode("utf8")



import json
import logging
import uuid

from django.core.paginator import InvalidPage, Paginator
from django.db import models
from django.http import Http404
from django.shortcuts import redirect
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel
from slugify import slugify
from wagtail.admin.edit_handlers import FieldPanel
from wagtail.core import blocks
from wagtail.core.fields import RichTextField, StreamField
from wagtail.core.models import Page, PageManager
from wagtail.embeds.blocks import EmbedBlock
from wagtail.images.blocks import ImageChooserBlock
from wagtail.images.models import Image
from wagtail.search import index

from cast import appsettings
from cast.blocks import AudioChooserBlock, GalleryBlock, VideoChooserBlock
from cast.filters import PostFilterset
from cast.models import get_or_create_gallery
from cast.models.itunes import ItunesArtWork

logger = logging.getLogger(__name__)


class Blog(TimeStampedModel, Page):
    author = models.CharField(max_length=255, default=None, null=True, blank=True)
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    email = models.EmailField(null=True, default=None, blank=True)
    comments_enabled = models.BooleanField(
        _("comments_enabled"),
        default=True,
        help_text=_("Whether comments are enabled for this blog." ""),
    )

    # podcast stuff

    # atm it's only used for podcast image
    itunes_artwork = models.ForeignKey(ItunesArtWork, null=True, blank=True, on_delete=models.SET_NULL)
    itunes_categories = models.CharField(
        _("itunes_categories"),
        max_length=512,
        blank=True,
        default="",
        help_text=_(
            "A json dict of itunes categories pointing to lists "
            "of subcategories. Taken from this list "
            "https://validator.w3.org/feed/docs/error/InvalidItunesCategory.html"
        ),
    )
    EXPLICIT_CHOICES = ((1, _("yes")), (2, _("no")), (3, _("clean")))
    keywords = models.CharField(
        _("keywords"),
        max_length=255,
        blank=True,
        default="",
        help_text=_(
            """A comma-delimitedlist of up to 12 words for iTunes
            searches. Perhaps include misspellings of the title."""
        ),
    )
    explicit = models.PositiveSmallIntegerField(
        _("explicit"),
        default=1,
        choices=EXPLICIT_CHOICES,
        help_text=_("``Clean`` will put the clean iTunes graphic by it."),
    )

    # wagtail
    description = RichTextField(blank=True)
    template = "cast/blog_list_of_posts.html"
    content_panels = Page.content_panels + [
        FieldPanel("description", classname="full"),
        FieldPanel("email"),
    ]

    subpage_types = ["cast.Post"]

    def __str__(self):
        return self.title

    @property
    def last_build_date(self):
        return Post.objects.live().descendant_of(self.blog).order_by("-visible_date")[0].visible_date

    @property
    def itunes_categories_parsed(self):
        try:
            return json.loads(self.itunes_categories)
        except json.decoder.JSONDecodeError:
            return {}

    @property
    def is_podcast(self):
        return Post.objects.live().descendant_of(self).exclude(podcast_audio__isnull=True).count() > 0

    @property
    def author_name(self):
        if self.author is not None:
            return self.author
        else:
            return self.owner.get_full_name()

    @property
    def unfiltered_published_posts(self):
        return Post.objects.live().descendant_of(self).order_by("-visible_date")

    @property
    def request(self):
        return getattr(self, "_request", None)

    @request.setter
    def request(self, value):
        self._request = value

    @property
    def filterset_data(self):
        if self.request is not None:
            return self.request.GET
        else:
            return getattr(self, "_filterset_data", {})

    @property
    def filterset(self):
        return PostFilterset(
            data=self.filterset_data, queryset=self.unfiltered_published_posts, fetch_facet_counts=True
        )

    @property
    def published_posts(self):
        return self.filterset.qs

    def paginate_queryset(self, context):
        paginator = Paginator(self.published_posts, appsettings.POST_LIST_PAGINATION)
        page = self.request.GET.get("page", False) or 1
        try:
            page_number = int(page)
        except ValueError:
            if page == "last":
                page_number = paginator.num_pages
            else:
                raise Http404(_("Page is not “last”, nor can it be converted to an int."))
        try:
            page = paginator.page(page_number)
        except InvalidPage as e:
            raise Http404(
                _("Invalid page (%(page_number)s): %(message)s") % {"page_number": page_number, "message": str(e)}
            )
        pagination_context = {
            "paginator": paginator,
            "page_obj": page,
            "is_paginated": page.has_other_pages(),
            "object_list": page.object_list,
        }
        context.update(pagination_context)
        return context

    def get_other_get_params(self):
        get_copy = self.request.GET.copy()
        parameters = get_copy.pop("page", True) and get_copy.urlencode()
        if len(parameters) > 0:
            parameters = f"&{parameters}"
        return parameters

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request, *args, **kwargs)
        self.request = request
        context["filterset"] = self.filterset
        context["parameters"] = self.get_other_get_params()
        context = self.paginate_queryset(context)
        context["posts"] = context["object_list"]  # convenience
        return context


class ContentBlock(blocks.StreamBlock):
    heading = blocks.CharBlock(classname="full title")
    paragraph = blocks.RichTextBlock()
    image = ImageChooserBlock(template="cast/image/image.html")
    gallery = GalleryBlock(ImageChooserBlock())
    embed = EmbedBlock()
    video = VideoChooserBlock(template="cast/video/video.html", icon="media")
    audio = AudioChooserBlock(template="cast/audio/audio.html", icon="media")

    class Meta:
        icon = "form"


def sync_media_ids(from_database, from_body):
    to_add, to_remove = {}, {}
    all_media_types = set(from_database.keys()).union(from_body.keys())
    for media_type in all_media_types:
        in_database_ids = from_database.get(media_type, set())
        in_body_ids = from_body.get(media_type, set())
        ids_to_add = in_body_ids - in_database_ids
        if len(ids_to_add) > 0:
            to_add[media_type] = ids_to_add
        ids_to_remove = in_database_ids - in_body_ids
        if len(ids_to_remove) > 0:
            to_remove[media_type] = ids_to_remove
    return to_add, to_remove


class PostPublishedManager(PageManager):
    use_for_related_fields = True

    def get_queryset(self):
        return super().get_queryset().filter(pub_date__lte=timezone.now())

    @property
    def podcast_episodes(self):
        return self.get_queryset().filter(podcast_audio__isnull=False)


class Post(TimeStampedModel, Page):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    pub_date = models.DateTimeField(null=True, blank=True)
    visible_date = models.DateTimeField(default=timezone.now)
    podcast_audio = models.ForeignKey(
        "cast.Audio", null=True, blank=True, on_delete=models.SET_NULL, related_name="posts"
    )
    keywords = models.CharField(
        _("keywords"),
        max_length=255,
        blank=True,
        default="",
        help_text=_(
            """A comma-demlimitedlist of up to 12 words for iTunes
            searches. Perhaps include misspellings of the title."""
        ),
    )
    explicit = models.PositiveSmallIntegerField(
        _("explicit"),
        choices=Blog.EXPLICIT_CHOICES,
        help_text=_("``Clean`` will put the clean iTunes graphic by it."),
        default=1,
    )
    block = models.BooleanField(
        _("block"),
        default=False,
        help_text=_(
            "Check to block this episode from iTunes because <br />its "
            "content might cause the entire show to be <br />removed from iTunes."
            ""
        ),
    )
    comments_enabled = models.BooleanField(
        _("comments_enabled"),
        default=True,
        help_text=_("Whether comments are enabled for this post." ""),
    )

    images = models.ManyToManyField(Image, blank=True)
    videos = models.ManyToManyField("cast.Video", blank=True)
    galleries = models.ManyToManyField("cast.Gallery", blank=True)
    audios = models.ManyToManyField("cast.Audio", blank=True)

    # wagtail
    body = StreamField(
        [
            ("overview", ContentBlock()),
            ("detail", ContentBlock()),
        ],
        use_json_field=True,
    )

    search_fields = Page.search_fields + [
        index.SearchField("body"),
    ]

    content_panels = Page.content_panels + [
        FieldPanel("visible_date"),
        FieldPanel("body"),
    ]
    template = "cast/post.html"
    parent_page_types = ["cast.Blog"]

    # managers
    objects = PageManager()
    published = PostPublishedManager()

    @property
    def media_model_lookup(self):
        from .audio import Audio
        from .gallery import Gallery
        from .video import Video

        return {
            "image": Image,
            "video": Video,
            "gallery": Gallery,
            "audio": Audio,
        }

    @property
    def blog(self):
        """
        The get_parent() method returns wagtail parent page, which is not
        necessarily a Blog model, but maybe the root page. If it's a Blog
        it has a .blog attribute containing the model which has all the
        attributes like blog.comments_enabled etc..
        """
        return self.get_parent().blog

    @property
    def is_published(self):
        return self.pub_date is not None and self.pub_date < timezone.now()

    def __str__(self):
        return self.title

    def get_enclosure_url(self, audio_format):
        return getattr(self.podcast_audio, audio_format).url

    def get_enclosure_size(self, audio_format):
        return getattr(self.podcast_audio, audio_format).size

    def get_slug(self):
        return slugify(self.title)

    @property
    def media_lookup(self):
        try:
            return {
                "image": {i.pk: i for i in self.images.all()},
                "video": {v.pk: v for v in self.videos.all()},
                "gallery": {g.pk: g for g in self.galleries.all()},
                "audio": {a.pk: a for a in self.audios.all()},
            }
        except ValueError:
            # post ist not yet saved
            pass

    @property
    def media_attr_lookup(self):
        return {
            "image": self.images,
            "video": self.videos,
            "gallery": self.galleries,
            "audio": self.audios,
        }

    @property
    def audio_in_body(self):
        audio_blocks = []
        for block in self.body:
            for content_block in block.value:
                if content_block.block_type == "audio":
                    audio_blocks.append(content_block)
        return len(audio_blocks) > 0

    @property
    def has_audio(self):
        audios_count = 0
        try:
            audios_count = self.audios.count()
        except ValueError:
            # will be raised on wagtail preview because page_ptr is not set
            pass
        return self.audio_in_body or audios_count > 0 or self.podcast_audio is not None

    @property
    def comments_are_enabled(self):
        return appsettings.CAST_COMMENTS_ENABLED and self.blog.comments_enabled and self.comments_enabled

    def get_context(self, *args, **kwargs):
        context = super().get_context(*args, **kwargs)
        context["render_detail"] = kwargs.get("render_detail", False)
        return context

    @property
    def media_ids_from_db(self):
        return {k: set(v) for k, v in self.media_lookup.items()}

    @property
    def media_ids_from_body(self):
        from_body = {}
        for content_block in self.body:
            for block in content_block.value:
                if block.block_type == "gallery":
                    image_ids = [i.id for i in block.value]
                    media_model = get_or_create_gallery(image_ids)
                else:
                    media_model = block.value
                if block.block_type in self.media_model_lookup:
                    from_body.setdefault(block.block_type, set()).add(media_model.id)
        return from_body

    def sync_media_ids(self):
        media_attr_lookup = self.media_attr_lookup
        to_add, to_remove = sync_media_ids(self.media_ids_from_db, self.media_ids_from_body)

        # add new ids
        for media_type, ids in to_add.items():
            for media_id in ids:
                media_attr_lookup[media_type].add(media_id)

        # remove obsolete ids
        for media_type, ids in to_remove.items():
            for media_id in ids:
                media_attr_lookup[media_type].remove(media_id)

    def save(self, *args, **kwargs):
        save_return = super().save(*args, **kwargs)
        self.sync_media_ids()
        return save_return


class HomePage(Page):
    body = StreamField(
        [
            ("heading", blocks.CharBlock(classname="full title")),
            ("paragraph", blocks.RichTextBlock()),
            ("image", ImageChooserBlock(template="cast/image/image.html")),
            ("gallery", GalleryBlock(ImageChooserBlock())),
        ],
        use_json_field=True,
    )
    alias_for_page = models.ForeignKey(
        "wagtailcore.Page",
        related_name="aliases_homepage",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
        verbose_name="Redirect to another page",
        help_text="Make this page an alias for another page, redirecting to it with a non permanent redirect.",
    )

    content_panels = Page.content_panels + [
        FieldPanel("alias_for_page"),
        FieldPanel("body"),
    ]

    def serve(self, request):
        if self.alias_for_page is not None:
            return redirect(self.alias_for_page.url, permanent=False)
        return super().serve(request)



from itertools import chain, islice, tee

from django.utils.functional import cached_property

from wagtail.core.blocks import ChooserBlock, ListBlock


def previous_and_next(iterable):
    prevs, items, nexts = tee(iterable, 3)
    prevs = chain([None], prevs)
    nexts = chain(islice(nexts, 1, None), [None])
    return zip(prevs, items, nexts)


class GalleryBlock(ListBlock):
    class Meta:
        template = "cast/gallery.html"

    def add_prev_next(self, gallery):
        for previous_image, current_image, next_image in previous_and_next(gallery):
            current_image.prev = "false" if previous_image is None else f"img-{previous_image.pk}"
            current_image.next = "false" if next_image is None else f"img-{next_image.pk}"

    def get_context(self, gallery, parent_context=None):
        self.add_prev_next(gallery)
        return super().get_context(gallery, parent_context=parent_context)


class VideoChooserBlock(ChooserBlock):
    @cached_property
    def target_model(self):
        from .models import Video

        return Video

    @cached_property
    def widget(self):
        from .widgets import AdminVideoChooser

        return AdminVideoChooser()

    def get_form_state(self, value):
        return self.widget.get_value_data(value)


class AudioChooserBlock(ChooserBlock):
    @cached_property
    def target_model(self):
        from .models import Audio

        return Audio

    @cached_property
    def widget(self):
        from .widgets import AdminAudioChooser

        return AdminAudioChooser()

    def get_form_state(self, value):
        return self.widget.get_value_data(value)



from django.conf import settings

CAST_COMMENTS_ENABLED = getattr(settings, "CAST_COMMENTS_ENABLED", False)
CHOOSER_PAGINATION = getattr(settings, "CHOOSER_PAGINATION", 10)
MENU_ITEM_PAGINATION = getattr(settings, "MENU_ITEM_PAGINATION", 20)
POST_LIST_PAGINATION = getattr(settings, "POST_LIST_PAGINATION", 5)
DELETE_WAGTAIL_IMAGES = getattr(settings, "DELETE_WAGTAIL_IMAGES", True)



from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

import pytest

from cast.models import Post


class TestPublished:
    pytestmark = pytest.mark.django_db

    def test_get_only_published_entries(self, client, unpublished_post):
        bp = unpublished_post
        feed_url = reverse("cast:latest_entries_feed", kwargs={"slug": bp.blog.slug})

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "xml" in content
        assert bp.title not in content

    def test_get_post_detail_not_published_not_auth(self, client, unpublished_post):
        post = unpublished_post
        detail_url = post.get_url()

        r = client.get(detail_url)
        assert r.status_code == 404

        content = r.content.decode("utf-8")
        assert post.title not in content

    def test_published_manager_pub_date_null(self, post):
        assert Post.published.count() == 1
        post.pub_date = None
        post.save()
        assert Post.objects.count() == 1
        assert Post.published.count() == 0

    def test_published_manager_future_pubdate(self, post):
        assert Post.published.count() == 1
        post.pub_date = timezone.now() + timedelta(seconds=10)
        post.save()
        assert Post.objects.count() == 1
        assert Post.published.count() == 0
        post.pub_date = timezone.now()
        post.save()
        assert Post.objects.count() == 1
        assert Post.published.count() == 1



from django.urls import include, path
from django.views.decorators.cache import cache_page

from . import feeds
from .views.dashboard import DashboardView

app_name = "cast"
urlpatterns = [
    # API
    # url(r"^api/", include("cast.api.urls", namespace="api")),
    path("api/", include("cast.api.urls", namespace="api")),
    # Dashboard
    path("dashboard/", view=DashboardView.as_view(), name="dashboard"),
    # Feeds
    path(
        "<slug:slug>/feed/rss.xml",
        view=cache_page(5 * 60)(feeds.LatestEntriesFeed()),
        name="latest_entries_feed",
    ),
    path(
        "<slug:slug>/feed/podcast/<audio_format>/rss.xml",
        view=cache_page(5 * 60)(feeds.RssPodcastFeed()),
        name="podcast_feed_rss",
    ),
    path(
        "<slug:slug>/feed/podcast/<audio_format>/atom.xml",
        view=cache_page(5 * 60)(feeds.AtomPodcastFeed()),
        name="podcast_feed_atom",
    ),
]



# from cast.access_log import get_last_request_position
# from cast.access_log import get_dataframe_from_position


# class TestParseAccesslog:
#    def test_get_access_log_position(self, access_log_path, last_request_dummy):
#        position = get_last_request_position(access_log_path, last_request_dummy)
#        assert position == 4
#
#    def test_get_df_from_access_log(self, access_log_path):
#        df = get_dataframe_from_position(access_log_path)
#        assert df.shape == (5, 9)



import string
from datetime import datetime
from itertools import chain

import django_filters
from django.db import models
from django.db.models.fields import BLANK_CHOICE_DASH
from django.forms import Widget
from django.forms.utils import flatatt
from django.utils.encoding import force_str
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _


class DateFacetWidget(Widget):
    def __init__(self, attrs=None, choices=()):
        super().__init__(attrs)

        self.choices = choices

    def value_from_datadict(self, data, files, name):
        value = super().value_from_datadict(data, files, name)
        self.data = data
        return value

    def render(self, name, value, attrs=None, choices=(), renderer=None):
        if not hasattr(self, "data"):
            self.data = {}
        if value is None:
            value = ""
        final_attrs = self.build_attrs(self.attrs, extra_attrs=attrs)
        output = ["<div%s>" % flatatt(final_attrs)]
        options = self.render_options(choices, [value], name)
        if options:
            output.append(options)
        output.append("</div>")
        return mark_safe("\n".join(output))

    def render_options(self, choices, selected_choices, name):
        selected_choices = {force_str(v) for v in selected_choices}
        output = []
        for option_value, option_label in chain(self.choices, choices):
            if isinstance(option_label, (list, tuple)):
                for option in option_label:
                    output.append(self.render_option(name, selected_choices, *option))
            else:
                output.append(self.render_option(name, selected_choices, option_value, option_label))
        return "\n".join(output)

    def render_option(self, name, selected_choices, option_value, option_label):
        option_value = force_str(option_value)
        if option_label == BLANK_CHOICE_DASH[0][1]:
            option_label = _("All")
        data = self.data.copy()
        data[name] = option_value
        selected = data == self.data or option_value in selected_choices
        try:
            url = data.urlencode()
        except AttributeError:
            url = urlencode(data)
        return self.option_string() % {
            "attrs": selected and ' class="selected"' or "",
            "query_string": url,
            "label": force_str(option_label),
        }

    def option_string(self):
        return '<div class="cast-date-facet-item"><a%(attrs)s href="?%(query_string)s">%(label)s</a></div>'


def parse_date_facets(value):
    """Split into function, because it needs to be imported by the post list view."""
    # clean up value a little bit, because otherwise sql-injection
    # search requests are spamming the logfile with garbage -> analytics wont work
    allowed = set(string.digits + "-")
    value = "".join([c for c in value if c in allowed])
    year_month = datetime.strptime(value, "%Y-%m")
    return year_month


def get_selected_facet(get_params):
    date_facet = get_params.get("date_facets")
    if date_facet is None or len(date_facet) == 0:
        return None
    return parse_date_facets(date_facet)


def get_facet_counts(filterset_data, queryset):
    filterset_data = {k: v for k, v in filterset_data.items()}  # copy filterset_data to avoid overwriting

    # get selected facet if set and build the facet counting queryset
    facet_counts = {}
    selected_facet = get_selected_facet(filterset_data)
    if selected_facet is not None:
        facet_counts = {"year_month": {selected_facet: 1}}
    filterset_data["facet_counts"] = facet_counts
    post_filter = PostFilterset(queryset=queryset, data=filterset_data, facet_counts=facet_counts)
    facet_queryset = (
        post_filter.qs.order_by()
        .annotate(month=models.functions.TruncMonth("visible_date"))
        .values("month")
        .annotate(n=models.Count("pk"))
    )

    # build up the date facet counts for final filter pass
    year_month_counts = {}
    for row in facet_queryset:
        year_month_counts[row["month"]] = row["n"]
    return {"year_month": year_month_counts}


class FacetChoicesMixin:
    """Just a way to pass the facet counts to the field which displays the choice."""

    @property
    def field(self):
        facet_count_choices = []
        for year_month, count in sorted(self.parent.facet_counts.get("year_month", {}).items()):
            date_str = year_month.strftime("%Y-%m")
            label = f"{date_str} ({count})"
            facet_count_choices.append((date_str, label))
        self.extra["choices"] = facet_count_choices
        return super().field


class DateFacetFilter(FacetChoicesMixin, django_filters.filters.ChoiceFilter):
    def filter(self, qs, value):
        if len(value) == 0:
            # don't filter if value is empty
            return qs
        year_month = parse_date_facets(value)
        year = year_month.year
        month = year_month.month
        return qs.filter(visible_date__year=year, visible_date__month=month)


class PostFilterset(django_filters.FilterSet):
    search = django_filters.CharFilter(field_name="title", method="fulltext_search", label="Search")
    date = django_filters.DateFromToRangeFilter(
        field_name="visible_date",
        label="Date",
        widget=django_filters.widgets.DateRangeWidget(attrs={"type": "date", "placeholder": "YYYY/MM/DD"}),
    )
    date_facets = DateFacetFilter(
        field_name="title",
        label="Date Facets",
        choices=[],
        widget=DateFacetWidget(attrs={"class": "cast-date-facet-container"}),
    )
    o = django_filters.OrderingFilter(
        fields=(("visible_date", "visible_date"),),
        field_labels={"visible_date": "Date"},
    )

    class Meta:
        fields = ["search", "date", "date_facets"]

    def __init__(
        self,
        data=None,
        queryset=None,
        *,
        facet_counts={},
        fetch_facet_counts=False,
    ):
        super().__init__(data=data, queryset=queryset)
        self.facet_counts = facet_counts
        if fetch_facet_counts:
            # avoid running into infinite recursion problems, because
            # get_facet_counts will instantiate Postfilterset again
            # -> and again -> and again ...
            try:
                self.facet_counts = get_facet_counts(data, queryset)
            except ValueError:
                self.facet_counts = {}

    @staticmethod
    def fulltext_search(queryset, name, value):
        return queryset.search(value).get_queryset()



from django.urls import path, re_path

from rest_framework.schemas import get_schema_view

from . import views


app_name = "api"
schema_view = get_schema_view(title="Cast API")

urlpatterns = [
    path("schema/", schema_view),
    path("", views.api_root, name="root"),
    # video
    path("videos/?", views.VideoListView.as_view(), name="video_list"),
    re_path(r"^videos/(?P<pk>\d+)/?$", views.VideoDetailView.as_view(), name="video_detail"),
    path(
        "upload_video/",
        views.VideoCreateView.as_view(),
        name="upload_video",
    ),
    # audio
    path("audio/?", views.AudioListView.as_view(), name="audio_list"),
    re_path(r"^audios/(?P<pk>\d+)/?$", views.AudioDetailView.as_view(), name="audio_detail"),
    re_path(
        r"^audios/podlove/(?P<pk>\d+)/?$",
        views.AudioPodloveDetailView.as_view(),
        name="audio_podlove_detail",
    ),
    # request
    path("request/?", views.RequestListView.as_view(), name="request_list"),
]



from datetime import timedelta

from django.urls import reverse

import pytest

from .factories import UserFactory


# from cast.access_log import pandas_rows_to_dict
# from cast.access_log import get_last_request_position
# from cast.access_log import get_dataframe_from_position


class TestBlogVideo:
    @classmethod
    def setup_class(cls):
        cls.list_url = reverse("cast:api:video_list")
        cls.detail_url = reverse("cast:api:video_detail", kwargs={"pk": 1})

    @pytest.mark.django_db
    def test_video_list_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the list
        endpoint without being authenticated.
        """
        r = api_client.get(self.list_url, format="json")
        assert r.status_code == 403

    @pytest.mark.django_db
    def test_video_detail_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the
        detail endpoint without being authenticated.
        """
        r = api_client.get(self.detail_url, format="json")
        assert r.status_code == 403

    @pytest.mark.django_db
    def test_video_list_endpoint_with_authentication(self, api_client):
        """Check for list result when accessing the list endpoint
        being logged in.
        """
        user = UserFactory()
        api_client.login(username=user.username, password="password")
        r = api_client.get(self.list_url, format="json")
        # dont redirect to login page
        assert r.status_code == 200
        assert "results" in r.json()


class TestBlogAudio:
    @classmethod
    def setup_class(cls):
        cls.list_url = reverse("cast:api:audio_list")
        cls.detail_url = reverse("cast:api:audio_detail", kwargs={"pk": 1})

    @pytest.mark.django_db
    def test_audio_list_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the list
        endpoint without being authenticated.
        """
        r = api_client.get(self.list_url, format="json")
        assert r.status_code == 403

    @pytest.mark.django_db
    def test_audio_detail_endpoint_without_authentication(self, api_client):
        """Check for not authenticated status code if trying to access the
        detail endpoint without being authenticated.
        """
        r = api_client.get(self.detail_url, format="json")
        assert r.status_code == 403

    @pytest.mark.django_db
    def test_audio_list_endpoint_with_authentication(self, api_client):
        """Check for list result when accessing the list endpoint
        being logged in.
        """
        user = UserFactory()
        api_client.login(username=user.username, password="password")
        r = api_client.get(self.list_url, format="json")
        # dont redirect to login page
        assert r.status_code == 200
        assert "results" in r.json()


class TestPodcastAudio:
    @pytest.mark.django_db
    def test_podlove_detail_endpoint_without_authentication(self, api_client, audio):
        """Should be accessible without authentication."""
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})
        r = api_client.get(podlove_detail_url, format="json")
        assert r.status_code == 200

    @pytest.mark.django_db
    def test_podlove_detail_endpoint_duration(self, api_client, audio):
        """Test whether microseconds get stripped away from duration via api - they have
        to be for podlove player to work.
        """
        delta = timedelta(days=0, hours=1, minutes=10, seconds=20, microseconds=40)
        audio.duration = delta
        audio.save()
        assert "." in str(audio.duration)
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})
        r = api_client.get(podlove_detail_url, format="json")
        assert "." not in r.json()["duration"]

    @pytest.mark.django_db
    def test_podlove_detail_endpoint_chaptermarks(self, api_client, audio, chaptermarks):
        """Test whether chaptermarks get delivered via podlove endpoint."""
        print("chaptermarks: ", chaptermarks)
        podlove_detail_url = reverse("cast:api:audio_podlove_detail", kwargs={"pk": audio.pk})
        r = api_client.get(podlove_detail_url, format="json")
        chapters = r.json()["chapters"]
        for chapter in chapters:
            # assert microseconds are stripped away
            assert "." not in chapter["start"]
            assert ":" in chapter["start"]
        assert len(chapters) == 3
        # assert reordering
        assert chapters[-1]["title"] == "coughing"


class TestRequest:
    @classmethod
    def setup_class(cls):
        cls.list_url = reverse("cast:api:request_list")

    @pytest.mark.django_db
    def test_request_list_endpoint_without_authentication(self, api_client):
        """Should not be accessible without authentication."""
        r = api_client.get(self.list_url, format="json")
        assert r.status_code == 403

    @pytest.mark.django_db
    def test_request_list_endpoint_with_authentication(self, api_client):
        """Check for list result when accessing the list endpoint
        being logged in.
        """
        user = UserFactory()
        api_client.login(username=user.username, password="password")
        r = api_client.get(self.list_url, format="json")
        # dont redirect to login page
        assert r.status_code == 200
        assert "results" in r.json()


#    @pytest.mark.django_db
#    def test_request_list_endpoint_non_bulk_insert(self, api_client, access_log_path):
#        user = UserFactory()
#        api_client.login(username=user.username, password="password")
#        df = get_dataframe_from_position(access_log_path, start_position=0)
#        raw_rows = df.fillna("").to_dict(orient="rows")
#        rows = pandas_rows_to_dict(raw_rows)
#        row = rows[0]
#        r = api_client.post(self.list_url, data=row, format="json")
#        assert r.status_code == 201

#    @pytest.mark.django_db
#    def test_request_list_endpoint_bulk_insert(self, api_client, access_log_path):
#        user = UserFactory()
#        api_client.login(username=user.username, password="password")
#        df = get_dataframe_from_position(access_log_path, start_position=0)
#        raw_rows = df.fillna("").to_dict(orient="rows")
#        rows = pandas_rows_to_dict(raw_rows)
#        r = api_client.post(self.list_url, data=rows, format="json")
#        assert r.status_code == 201
#        assert Request.objects.count() == df.shape[0]

#    @pytest.mark.django_db
#    def test_request_list_endpoint_incremental_insert(
#        self, api_client, access_log_path
#    ):
#        user = UserFactory()
#        api_client.login(username=user.username, password="password")
#        Request.objects.all().delete()
#
#        # insert just first row
#        df = get_dataframe_from_position(access_log_path, start_position=0)
#        raw_rows = df.fillna("").to_dict(orient="rows")
#        rows = pandas_rows_to_dict(raw_rows)
#        row = rows[0]
#        r = api_client.post(self.list_url, data=row, format="json")
#        assert r.status_code == 201
#
#        # get last position (should be 4 because first 5 are the same)
#        last_request = Request.objects.all().order_by("-timestamp")[0]
#        last_position = get_last_request_position(access_log_path, last_request)
#        assert last_position == 4
#
#        # insert starting at position 4
#        df = get_dataframe_from_position(access_log_path, start_position=last_position)
#        raw_rows = df.fillna("").to_dict(orient="rows")
#        rows = pandas_rows_to_dict(raw_rows)
#        r = api_client.post(self.list_url, data=rows, format="json")
#        assert r.status_code == 201
#
#        # assert number of unique lines in access.log and objects in database are equal
#        # we omitted some lines intentionally
#        number_of_unique_lines = 0
#        with open(access_log_path) as f:
#            number_of_unique_lines = len(set([l for l in f]))
#        assert Request.objects.count() == number_of_unique_lines



import json
import re
from collections import defaultdict

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from model_utils.models import TimeStampedModel

token_pattern = re.compile(r"(?u)\b\w\w+\b")
standard_tokenizer = token_pattern.findall


def regex_tokenize(message):
    return standard_tokenizer(message.lower())


def normalize(probabilities):
    try:
        factor = 1.0 / float(sum(probabilities.values()))
    except ZeroDivisionError:
        # not possible to scale -> skip
        return probabilities
    for name, value in probabilities.items():
        probabilities[name] *= factor
    return probabilities


class NaiveBayes:
    def __init__(self, tokenize=regex_tokenize, prior_probabilities={}, word_label_counts=None):
        self.tokenize = tokenize
        self.prior_probabilities = prior_probabilities
        if word_label_counts is None:
            self.word_label_counts = defaultdict(lambda: defaultdict(int))
        else:
            self.word_label_counts = word_label_counts
        self.number_of_words = self.get_number_of_words(self.word_label_counts)

    @staticmethod
    def get_label_counts(messages):
        label_counts = defaultdict(int)
        for label, text in messages:
            label_counts[label] += 1
        return label_counts

    def set_prior_probabilities(self, label_counts):
        number_of_messages = sum(label_counts.values())
        self.prior_probabilities = {label: count / number_of_messages for label, count in label_counts.items()}

    def set_word_label_counts(self, messages):
        counts = self.word_label_counts
        for label, text in messages:
            for word in self.tokenize(text):
                counts[word][label] += 1

    @staticmethod
    def get_number_of_words(word_label_counts):
        number_of_words = defaultdict(int)
        for word, counts in word_label_counts.items():
            for label, count in counts.items():
                number_of_words[label] += 1
        return number_of_words

    def fit(self, messages):
        self.set_prior_probabilities(self.get_label_counts(messages))
        self.set_word_label_counts(messages)
        self.number_of_words = self.get_number_of_words(self.word_label_counts)
        return self

    @staticmethod
    def update_probabilities(probabilities, counts_per_label, number_of_words):
        updated_probabilities = {}
        for label, prior_probability in probabilities.items():
            word_count = counts_per_label.get(label, 0.5)
            word_probability = word_count / number_of_words[label]
            updated_probabilities[label] = prior_probability * word_probability
        return updated_probabilities

    def predict(self, message):
        probabilities = dict(self.prior_probabilities)
        for word in self.tokenize(message):
            counts_per_label = self.word_label_counts.get(word, {})
            probabilities = normalize(self.update_probabilities(probabilities, counts_per_label, self.number_of_words))
        return probabilities

    def predict_label(self, message):
        probabilities = self.predict(message)
        if len(probabilities) == 0:
            return None
        return sorted(((prob, label) for label, prob in probabilities.items()), reverse=True)[0][1]

    def dict(self):
        return {
            "class": "NaiveBayes",
            "prior_probabilities": self.prior_probabilities,
            "word_label_counts": self.word_label_counts,
        }

    def __eq__(self, other):
        return (
            self.prior_probabilities == other.prior_probabilities and self.word_label_counts == other.word_label_counts
        )


class ModelEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, NaiveBayes):
            return obj.dict()
        return super().default(obj)


class ModelDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        kwargs["object_hook"] = self.model_decode
        super().__init__(*args, **kwargs)

    @staticmethod
    def model_decode(obj):
        if obj.get("class") == "NaiveBayes":
            del obj["class"]
            return NaiveBayes(**obj)
        return obj


class SpamFilter(TimeStampedModel):
    name = models.CharField(unique=True, max_length=128)
    model = models.JSONField(verbose_name="Spamfilter Model", default=dict, encoder=ModelEncoder, decoder=ModelDecoder)

    @classmethod
    def comment_to_message(cls, comment):
        return f"{comment.name} {comment.title} {comment.comment}"

    @classmethod
    def get_training_data_comments(cls):
        from django_comments import get_model as get_comments_model

        Comment = get_comments_model()
        train = []
        for comment in Comment.objects.all():
            label = "ham" if (comment.is_public and not comment.is_removed) else "spam"
            message = cls.comment_to_message(comment)
            train.append((label, message))
        return train

    def retrain_from_scratch(self):
        """
        Retrain on all comments for now. Later on there might be
        different spamfilters for different blogs/sites..
        """
        train = SpamFilter.get_training_data_comments()
        model = NaiveBayes().fit(train)
        self.model = model
        self.save()

    @classmethod
    @property
    def default(cls):
        return cls.objects.first()



from django.urls import include, path, reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from wagtail.admin.menu import MenuItem
from wagtail.core import hooks
from wagtail.core.permission_policies.collections import (
    CollectionOwnershipPermissionPolicy,
)

from .admin_urls import audio, video
from .models import Audio, Video


@hooks.register("register_admin_urls")
def register_admin_urls():
    return [
        path("audio/", include((audio, "castaudio"), namespace="castaudio")),
        path("media/", include((video, "castvideo"), namespace="castvideo")),
    ]


class VideoMenuItem(MenuItem):
    def is_shown(self, request):
        permission_policy = CollectionOwnershipPermissionPolicy(Video, auth_model=Video, owner_field_name="user")
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_video_menu_item():
    return VideoMenuItem(
        _("Video"),
        reverse("castvideo:index"),
        name="video",
        classnames="icon icon-media",
        order=300,
    )


@hooks.register("insert_editor_js")
def editor_js():
    return format_html(
        """
        <script>
            window.chooserUrls.videoChooser = '{0}';
        </script>
        """,
        reverse("castvideo:chooser"),
    )


class AudioMenuItem(MenuItem):
    def is_shown(self, request):
        permission_policy = CollectionOwnershipPermissionPolicy(Audio, auth_model=Audio, owner_field_name="user")
        return permission_policy.user_has_any_permission(request.user, ["add", "change", "delete"])


@hooks.register("register_admin_menu_item")
def register_audio_menu_item():
    return AudioMenuItem(
        _("Audio"),
        reverse("castaudio:index"),
        name="audio",
        classnames="icon icon-media",
        order=300,
    )


@hooks.register("insert_editor_js")
def editor_js_audio():
    return format_html(
        """
        <script>
            window.chooserUrls.audioChooser = '{0}';
        </script>
        """,
        reverse("castaudio:chooser"),
    )



import logging

from django.http import HttpResponse


logger = logging.getLogger(__name__)


class FileUploadResponseMixin:
    def get_success_url(self):
        return None

    def form_valid(self, form):
        model = form.save(commit=False)
        super().form_valid(form)
        return HttpResponse(f"{model.pk}", status=201)


class AddRequestUserMixin:
    user_field_name = "user"

    def form_valid(self, form):
        model = form.save(commit=False)
        setattr(model, self.user_field_name, self.request.user)
        return super().form_valid(form)



import pytest

from cast.models.moderation import NaiveBayes, SpamFilter, normalize


@pytest.mark.parametrize(
    "train, expected_initial_probabilities",
    [
        ([], {}),
        ([("ham", ""), ("spam", "")], {"ham": 0.5, "spam": 0.5}),
        ([(i, "") for i in range(3)], {k: 1 / 3 for k in range(3)}),
    ],
)
def test_initial_probabilities(train, expected_initial_probabilities):
    model = NaiveBayes().fit(train)
    assert model.predict("") == expected_initial_probabilities


@pytest.mark.parametrize(
    "probabilities, counts_per_label, number_of_words, expected",
    [
        ({}, {}, {}, {}),
        ({"spam": 1}, {"spam": 10}, {"spam": 100}, {"spam": 0.1}),
        (
            {"spam": 1, "ham": 0.5},
            {"spam": 10, "ham": 5},
            {"spam": 100, "ham": 50},
            {"spam": 0.1, "ham": 0.05},
        ),
        # no word count -> make sure probability is > 0
        ({"spam": 1}, {}, {"spam": 100}, {"spam": 0.005}),
    ],
)
def test_predict_word_probabilities(probabilities, counts_per_label, number_of_words, expected):
    model = NaiveBayes()
    updated = model.update_probabilities(probabilities, counts_per_label, number_of_words)
    assert updated == expected


@pytest.mark.parametrize(
    "message, expected_probabilities",
    [("foo", {"ham": (1 / 3) * (0.5 / 2), "spam": (2 / 3) * (2 / 5)})],
)
def test_predict(message, expected_probabilities):
    train = [
        ("spam", "foo bar baz"),
        ("spam", "foo asdf bsdf"),
        ("ham", "asdf csdf"),
    ]
    model = NaiveBayes().fit(train)
    probabilities = model.predict(message)
    assert probabilities == normalize(expected_probabilities)


@pytest.mark.parametrize(
    "train, message, expected_label",
    [
        ([], "asdf", None),
        ([("spam", "foo bar baz")], "asdf", "spam"),
    ],
)
def test_predict_label(train, message, expected_label):
    model = NaiveBayes().fit(train)
    label = model.predict_label(message)
    assert label == expected_label


@pytest.mark.django_db()
def test_model_default_serialization():
    class StubModel:
        foo = "blub"

    spamfilter = SpamFilter(name="stub model", model=StubModel())
    with pytest.raises(TypeError):
        # StubModel is not json serializable -> TypeError
        # makes sure return super().default(obj) for ModelEncoder is called
        spamfilter.save()


@pytest.mark.django_db()
def test_model_naive_bayes_serialization():
    train = [
        ("spam", "foo bar baz"),
        ("ham", "asdf bsdf csdf"),
    ]
    model = NaiveBayes().fit(train)
    spamfilter = SpamFilter(name="naive bayes", model=model)
    spamfilter.save()

    spamfilter.refresh_from_db()
    model_from_db = spamfilter.model
    assert model_from_db == model


@pytest.mark.django_db()
def test_spamfilter_retrain_from_scratch(comment, comment_spam):
    """
    Make sure data from comments is used when retraining
    spamfilter from scratch.
    """
    model = NaiveBayes().fit([])
    spamfilter = SpamFilter(name="naive bayes", model=model)
    spamfilter.save()
    assert spamfilter.model.prior_probabilities == {}

    spamfilter.retrain_from_scratch()
    assert spamfilter.model.prior_probabilities == {"ham": 0.5, "spam": 0.5}



import logging

from collections import OrderedDict

from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse
from django.views.generic import CreateView

from rest_framework import generics, status
from rest_framework.decorators import api_view
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.serializers import ListSerializer

from ..forms import VideoForm
from ..models import Audio, Request, Video
from .serializers import (
    AudioPodloveSerializer,
    AudioSerializer,
    RequestSerializer,
    VideoSerializer,
)
from .viewmixins import AddRequestUserMixin, FileUploadResponseMixin


logger = logging.getLogger(__name__)


@api_view(["GET"])
def api_root(request):
    """
    Show API contents.
    If you add any object types, add them here!
    """
    root_api_urls = (
        ("images", request.build_absolute_uri(reverse("cast:api:image_list"))),
        ("galleries", request.build_absolute_uri(reverse("cast:api:gallery_list"))),
        ("videos", request.build_absolute_uri(reverse("cast:api:video_list"))),
        ("audios", request.build_absolute_uri(reverse("cast:api:audio_list"))),
        ("requests", request.build_absolute_uri(reverse("cast:api:request_list"))),
    )
    return Response(OrderedDict(root_api_urls))


class VideoCreateView(LoginRequiredMixin, AddRequestUserMixin, FileUploadResponseMixin, CreateView):
    model = Video
    form_class = VideoForm
    user_field_name = "user"


class StandardResultsSetPagination(PageNumberPagination):
    page_size = 40
    page_size_query_param = "pageSize"
    max_page_size = 10000


class VideoListView(generics.ListCreateAPIView):
    serializer_class = VideoSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        user = self.request.user
        qs = Video.objects.all().filter(user=user)
        return qs.order_by("-created")


class VideoDetailView(generics.RetrieveDestroyAPIView):
    queryset = Video.objects.all()
    serializer_class = VideoSerializer
    permission_classes = (IsAuthenticated,)


class AudioListView(generics.ListCreateAPIView):
    serializer_class = AudioSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def get_queryset(self):
        user = self.request.user
        qs = Audio.objects.all().filter(user=user)
        return qs.order_by("-created")


class AudioDetailView(generics.RetrieveDestroyAPIView):
    queryset = Audio.objects.all()
    serializer_class = AudioSerializer
    permission_classes = (IsAuthenticated,)


class AudioPodloveDetailView(generics.RetrieveAPIView):
    queryset = Audio.objects.all()
    serializer_class = AudioPodloveSerializer


class RequestListView(generics.ListCreateAPIView):
    queryset = Request.objects.all().order_by("-timestamp")
    serializer_class = RequestSerializer
    pagination_class = StandardResultsSetPagination
    permission_classes = (IsAuthenticated,)

    def create(self, request, *args, **kwargs):
        """Allow for bulk create via many=True."""
        serializer = self.get_serializer(data=request.data, many=isinstance(request.data, list))
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create(self, serializer):
        """Use bulk_create for request lists, normal model serializer otherwise."""
        if isinstance(serializer, ListSerializer):
            requests = [Request(**d) for d in serializer.validated_data]
            Request.objects.bulk_create(requests)
        else:
            serializer.save()



from django.db import models


class Request(models.Model):
    """
    Hold requests from access.log files.
    """

    ip = models.GenericIPAddressField()
    timestamp = models.DateTimeField()
    status = models.PositiveSmallIntegerField()
    size = models.PositiveIntegerField()
    referer = models.CharField(max_length=2048, blank=True, null=True)
    user_agent = models.CharField(max_length=1024, blank=True, null=True)

    REQUEST_METHOD_CHOICES = [
        (1, "GET"),
        (2, "HEAD"),
        (3, "POST"),
        (4, "PUT"),
        (5, "PATCH"),
        (6, "DELETE"),
        (7, "OPTIONS"),
        (8, "CONNECT"),
        (8, "TRACE"),
    ]
    method = models.PositiveSmallIntegerField(choices=REQUEST_METHOD_CHOICES)

    path = models.CharField(max_length=1024)

    HTTP_PROTOCOL_CHOICES = [(1, "HTTP/1.0"), (2, "HTTP/1.1"), (3, "HTTP/2.0")]
    protocol = models.PositiveSmallIntegerField(choices=HTTP_PROTOCOL_CHOICES)

    def __str__(self):
        return f"{self.pk} {self.ip} {self.path}"



from django.urls import include, re_path

from wagtail.admin import urls as wagtailadmin_urls
from wagtail.core import urls as wagtail_urls
from wagtail.documents import urls as wagtaildocs_urls

from rest_framework.documentation import include_docs_urls


urlpatterns = [
    # allauth
    re_path(r"^accounts/", include("allauth.urls")),
    # rest framework docs/schema urls
    re_path(r"^docs/", include_docs_urls(title="cast API service")),
    re_path(r"^cast/", include("cast.urls", namespace="cast")),
    # comments
    re_path(r"^posts/comments/", include("fluent_comments.urls")),
    # wagtail
    re_path(r"^cms/", include(wagtailadmin_urls)),
    re_path(r"^documents/", include(wagtaildocs_urls)),
    re_path(r"", include(wagtail_urls)),  # default is wagtail
]



import logging

from django.contrib import admin

from .models import (  # Image,
    Audio,
    Blog,
    ChapterMark,
    File,
    Gallery,
    ItunesArtWork,
    Post,
    SpamFilter,
    Video,
)


logger = logging.getLogger(__name__)


class AdminUserMixin:
    def get_changeform_initial_data(self, request):
        return {"user": request.user, "author": request.user}


class BlogModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("title", "owner")


admin.site.register(Blog, BlogModelAdmin)


class PostModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("title", "owner", "blog")


admin.site.register(Post, PostModelAdmin)


class ItunesArtWorkModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("pk", "original")
    fields = ("original",)


admin.site.register(ItunesArtWork, ItunesArtWorkModelAdmin)


class FileModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("original", "user")
    fields = ("user", "original")


admin.site.register(File, FileModelAdmin)


class AudioAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("user", "title", "subtitle", "m4a", "mp3", "oga", "opus")
    fields = ("user", "title", "subtitle", "m4a", "mp3", "oga", "opus")


admin.site.register(Audio, AudioAdmin)


class ChapterMarkModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("start", "title", "link", "image", "audio")


admin.site.register(ChapterMark, ChapterMarkModelAdmin)


class VideoModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("pk", "user")

    def save_model(self, request, obj, form, change):
        logger.info(f"poster: {obj.poster}")
        logger.info(f"form: {form.cleaned_data}")
        if change and not form.cleaned_data["poster"]:
            logger.info("poster was cleared")
            obj.calc_poster = False
        super().save_model(request, obj, form, change)


admin.site.register(Video, VideoModelAdmin)


class GalleryModelAdmin(AdminUserMixin, admin.ModelAdmin):
    list_display = ("pk",)
    fields = ("user", "images")


admin.site.register(Gallery, GalleryModelAdmin)


@admin.action(description="Retrain model from scratch using marked comments")
def retrain(modeladmin, request, queryset):
    for spamfilter in queryset:
        spamfilter.retrain_from_scratch()


class SpamfilterModelAdmin(admin.ModelAdmin):
    list_display = ("pk", "name")
    fields = ("name",)
    actions = [retrain]


admin.site.register(SpamFilter, SpamfilterModelAdmin)



from django.urls import path

from ..views import audio as views


urlpatterns = [
    path("audio", views.index, name="index"),
    path(r"audio/add/", views.add, name="add"),
    path("audio/edit/<int:audio_id>/", views.edit, name="edit"),
    path("audio/delete/<int:audio_id>/", views.delete, name="delete"),
    path("audio/chooser/", views.chooser, name="chooser"),
    path("audio/chooser/upload/", views.chooser_upload, name="chooser_upload"),
    path("audio/chooser/<int:audio_id>/", views.chosen, name="chosen"),
]



from datetime import datetime
from time import mktime

import feedparser
import pytest
import pytz
from django.http import Http404
from django.urls import reverse

from cast.feeds import ITunesElements, PodcastFeed
from cast.models import Post


class TestFeedCreation:
    @pytest.mark.django_db
    def test_add_artwork_true(self, dummy_handler, blog_with_artwork):
        ie = ITunesElements()
        ie.feed = {"title": "foobar", "link": "bar"}
        ie.add_artwork(blog_with_artwork, dummy_handler)
        assert "itunes:image" in dummy_handler.aqe
        assert "image" in dummy_handler.se
        assert "image" in dummy_handler.ee

    @pytest.mark.django_db
    def test_add_artwork_false(self, dummy_handler, blog):
        ie = ITunesElements()
        ie.feed = {"title": "foobar", "link": "bar"}
        ie.add_artwork(blog, dummy_handler)
        assert "itunes:image" not in dummy_handler.aqe
        assert "image" not in dummy_handler.se
        assert "image" not in dummy_handler.ee

    def test_unknown_audio_format(self):
        pf = PodcastFeed()
        with pytest.raises(Http404):
            pf.set_audio_format("foobar")

    @pytest.mark.django_db
    def test_itunes_categories(self, dummy_handler, blog_with_itunes_categories):
        blog = blog_with_itunes_categories
        ie = ITunesElements()
        ie.add_itunes_categories(blog, dummy_handler)
        assert dummy_handler.se["itunes:category"]["text"] == "foo"
        assert dummy_handler.aqe["itunes:category"][-1]["text"] == "baz"
        assert "itunes:category" in dummy_handler.ee


@pytest.fixture
def use_dummy_cache_backend(settings):
    settings.CACHES = {
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        }
    }


class TestGeneratedFeeds:
    @pytest.mark.django_db
    def test_get_latest_entries_feed(self, client, post, use_dummy_cache_backend):
        feed_url = reverse("cast:latest_entries_feed", kwargs={"slug": post.blog.slug})

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "xml" in content
        assert post.title in content

    @pytest.mark.django_db
    def test_get_podcast_m4a_feed_rss(self, client, podcast_episode, use_dummy_cache_backend):
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": podcast_episode.blog.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "rss" in content
        assert podcast_episode.title in content

    @pytest.mark.django_db
    def test_get_podcast_m4a_feed_atom(self, client, podcast_episode):
        feed_url = reverse(
            "cast:podcast_feed_atom",
            kwargs={"slug": podcast_episode.blog.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        content = r.content.decode("utf-8")
        assert "feed" in content
        assert podcast_episode.title in content

    @pytest.mark.django_db
    def test_podcast_feed_contains_only_podcasts(self, client, post, podcast_episode, use_dummy_cache_backend):
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": podcast_episode.blog.slug, "audio_format": "m4a"},
        )
        r = client.get(feed_url)
        assert r.status_code == 200

        d = feedparser.parse(r.content)
        assert len(d.entries) == 1
        assert Post.objects.live().descendant_of(podcast_episode.blog).count() == 2

    @pytest.mark.django_db
    def test_podcast_feed_contains_visible_date_as_pubdate(
        self, client, podcast_episode_with_different_visible_date, use_dummy_cache_backend
    ):
        podcast_episode = podcast_episode_with_different_visible_date
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": podcast_episode.blog.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        d = feedparser.parse(r.content)
        date_from_feed = datetime.fromtimestamp(mktime(d.entries[0]["published_parsed"]))
        date_from_feed = pytz.utc.localize(date_from_feed)
        assert date_from_feed == podcast_episode.visible_date

    @pytest.mark.django_db
    def test_podcast_feed_contains_detail_information(self, client, podcast_episode):
        feed_url = reverse(
            "cast:podcast_feed_rss",
            kwargs={"slug": podcast_episode.blog.slug, "audio_format": "m4a"},
        )

        r = client.get(feed_url)
        assert r.status_code == 200

        d = feedparser.parse(r.content)
        content = d.entries[0]["content"][0]["value"]
        assert "in_all" in content
        assert "only_in_detail" in content



from datetime import timedelta

import pytest

from cast.models import Audio, ChapterMark, sync_chapter_marks


class TestVideoModel:
    @pytest.mark.django_db
    def test_get_all_video_paths(self, video):
        all_paths = list(video.get_all_paths())
        assert len(all_paths) == 1

    @pytest.mark.django_db
    def test_get_all_video_paths_with_poster(self, video_with_poster):
        all_paths = list(video_with_poster.get_all_paths())
        assert len(all_paths) == 2

    @pytest.mark.django_db
    def test_get_all_video_paths_without_thumbnail(self, video):
        class Dummy:
            name = "foobar"
            closed = True

            def open(self):
                return None

            def close(self):
                return None

            def seek(self, position):
                return None

            def read(self, num_bytes):
                return b""

            def tell(self):
                return 0

        video.poster = Dummy()
        all_paths = list(video.get_all_paths())
        assert len(all_paths) == 2


class TestGalleryModel:
    @pytest.mark.django_db
    def test_get_image_ids(self, gallery):
        assert len(gallery.image_ids) == gallery.images.count()


class TestAudioModel:
    @pytest.mark.django_db
    def test_get_file_formats(self, audio):
        assert audio.file_formats == "m4a"

    @pytest.mark.django_db
    def test_get_file_names(self, audio):
        assert "test" in audio.get_audio_file_names()

    @pytest.mark.django_db
    def test_get_name(self, audio):
        audio.title = None  # make sure name is provided by file
        assert audio.name == "test"

    @pytest.mark.django_db
    def test_get_name_with_title(self, audio):
        title = "foobar"
        audio.title = title
        assert audio.name == title

    @pytest.mark.django_db
    def test_audio_str(self, audio):
        audio.title = None  # make sure name is provided by file
        assert "1 - test" == str(audio)

    @pytest.mark.django_db
    def test_audio_get_all_paths(self, audio):
        assert "cast_audio/test.m4a" in audio.get_all_paths()

    @pytest.mark.django_db
    def test_audio_duration(self, audio):
        duration = audio._get_audio_duration(audio.m4a.path)
        assert duration == timedelta(microseconds=700000)

    @pytest.mark.django_db
    def test_audio_create_duration(self, audio):
        duration = "00:01:01.00"
        audio._get_audio_duration = lambda x: duration
        audio.create_duration()
        assert audio.duration == duration

    @pytest.mark.django_db
    def test_audio_podlove_url(self, audio):
        assert audio.podlove_url == "/cast/api/audios/podlove/1"

    @pytest.mark.django_db
    def test_audio_get_chaptermark_data_from_file_empty_on_value_error(self, audio):
        assert audio.get_chaptermark_data_from_file("mp3") == []

    def test_audio_test_clean_ffprobe_chaptermarks(self):
        ffprobe_chaptermarks = {
            "chapters": [
                {
                    "id": 0,
                    "time_base": "1/1000",
                    "start": 0,
                    "start_time": "0.000000",
                    "end": 155343,
                    "end_time": "155.343000",
                    # chapter marks with empty title should be filtered
                    "tags": {"title": ""},
                },
                {
                    "id": 1,
                    "time_base": "1/1000",
                    "start": 155343,
                    "start_time": "155.343000",
                    "end": 617117,
                    "end_time": "617.117000",
                    "tags": {"title": "News aus der Szene"},
                },
                {
                    "id": 2,
                    "time_base": "1/1000",
                    "start": 617117,
                    "start_time": "617.117000",
                    "end": 1266062,
                    "end_time": "1266.062000",
                    "tags": {"title": "Django Async"},
                },
            ]
        }
        cleaned_chaptermarks = Audio.clean_ffprobe_chaptermarks(ffprobe_chaptermarks)
        assert len(cleaned_chaptermarks) == 2
        assert cleaned_chaptermarks == [
            {"start": "155.343000", "title": "News aus der Szene"},
            {"start": "617.117000", "title": "Django Async"},
        ]


class TestFileModel:
    @pytest.mark.django_db
    def test_get_all_file_paths(self, file_instance):
        all_paths = list(file_instance.get_all_paths())
        assert len(all_paths) == 1


class TestBlogModel:
    @pytest.mark.django_db
    def test_blog_str(self, blog):
        assert blog.title == str(blog)

    @pytest.mark.django_db
    def test_blog_author_null(self, blog):
        blog.author = None
        assert blog.author_name == blog.owner.get_full_name()

    @pytest.mark.django_db
    def test_blog_author_not_null(self, blog):
        blog.author = "Foobar"
        assert blog.author_name == blog.author


class TestPostModel:
    @pytest.mark.django_db
    def test_post_slug(self, post):
        assert post.get_slug() == "test-entry"

    @pytest.mark.django_db
    def test_post_has_audio(self, post):
        assert post.has_audio is False

    @pytest.mark.django_db
    def test_post_has_audio_true(self, post, audio):
        post.podcast_audio = audio
        assert post.has_audio is True

    @pytest.mark.django_db
    def test_post_comments_enabled(self, post, comments_enabled):
        post.comments_enabled = True
        post.blog.comments_enabled = True
        assert post.comments_are_enabled

    @pytest.mark.django_db
    def test_post_comments_disabled_settings(self, post, comments_not_enabled):
        post.comments_enabled = True
        post.blog.comments_enabled = True
        assert not post.comments_are_enabled

    @pytest.mark.django_db
    def test_post_comments_disabled_blog(self, post, comments_enabled):
        post.comments_enabled = True
        post.blog.comments_enabled = False
        assert not post.comments_are_enabled

    @pytest.mark.django_db
    def test_post_comments_disabled_post(self, post, comments_enabled):
        post.comments_enabled = False
        post.blog.comments_enabled = True
        assert not post.comments_are_enabled


class TestChapterMarkModel:
    pytestmark = pytest.mark.django_db

    def test_chaptermark_original_line(self, chaptermarks):
        chaptermark = chaptermarks[0]
        assert chaptermark.original_line == "00:01:01.234 introduction  "

    def test_chaptermark_original_line_link(self, chaptermarks):
        link = "http://foobar.com"
        chaptermark = chaptermarks[0]
        chaptermark.link = link
        assert chaptermark.original_line == f"00:01:01.234 introduction {link} "

    def test_chaptermark_original_line_image(self, chaptermarks):
        image = "http://foobar.com/blub.jpg"
        chaptermark = chaptermarks[0]
        chaptermark.image = image
        assert chaptermark.original_line == f"00:01:01.234 introduction  {image}"


def start_strings_to_chaptermarks(audio, start_strings):
    chaptermarks = []
    for start_string in start_strings:
        cm = ChapterMark(audio=audio, start=start_string, title="foobar")
        cm.full_clean()  # convert string to datetime.time
        chaptermarks.append(cm)
    return chaptermarks


def chaptermarks_are_equal(actual, expected):
    if len(actual) != len(expected):
        return False
    attributes_to_compare = ["start", "title"]
    for a, b in zip(actual, expected):
        for attr in attributes_to_compare:
            if getattr(a, attr) != getattr(b, attr):
                return False
    return True


@pytest.mark.parametrize(
    "from_database, from_cms, expected_to_add, expected_to_update, expected_to_remove",
    [
        # from_database, from_cms, expected_to_add, expected_to_update, expected_to_remove
        ([], [], [], [], []),
        ([], ["0:0"], ["0:0"], [], []),  # add chaptermark
        (["0:0"], ["0:0"], [], [], []),  # update chaptermark -> empty because equal
        (["0:0"], [], [], [], ["0:0"]),  # remove chaptermark
        # add, update and remove together
        (["0:0", "0:2"], ["0:0", "0:1"], ["0:1"], [], ["0:2"]),
    ],
)
@pytest.mark.django_db
def test_sync_chapter_marks(from_database, from_cms, expected_to_add, expected_to_update, expected_to_remove, audio):
    args = []
    for start_string_list in [from_database, from_cms, expected_to_add, expected_to_update, expected_to_remove]:
        args.append(start_strings_to_chaptermarks(audio, start_string_list))
    from_database, from_cms, expected_to_add, expected_to_update, expected_to_remove = args
    actual_to_add, actual_to_update, actual_to_remove = sync_chapter_marks(from_database, from_cms)
    for actual, expected in zip(
        [actual_to_add, actual_to_update, actual_to_remove], [expected_to_add, expected_to_update, expected_to_remove]
    ):
        assert chaptermarks_are_equal(actual, expected)


@pytest.mark.django_db
def test_sync_chapter_marks_update_changed_chaptermarks(audio):
    cm1 = ChapterMark(audio=audio, start="0:0", title="foobar")
    cm1.full_clean()
    cm2 = ChapterMark(audio=audio, start="0:0", title="changed")
    cm2.full_clean()
    actual_to_add, actual_to_update, actual_to_remove = sync_chapter_marks([cm1], [cm2])
    assert len(actual_to_update) == 1
